-- File name   : EFX_VOXEL.MS
-- Title       : Voxel Generator
-- Category    : Geometry & Helper
-- Environment : 3ds max 3&4
-- Description : 3dsmax to txt exporter (for use with XCC mixer to convert to voxel format)
-- Reference   :
-- version	   : 1.02
-- engine Copyright (c) 2002, Will Sutton
-- game adaptation : Geert Van der Poorten

-- global variables -----------------------------------------------------------

global efx_floater_voxel
global efx_rollout_voxel_about
global efx_rollout_voxel_param
vxlcount =0
vxlcreated =0
vxlsize=0
-- general functions ----------------------------------------------------------
-- cen : ceter point
-- ext : size

backgroundColor = color 0 0 255 
setRefCoordSys #world 


-- global variables -----------------------------------------------------------

global sec -- body array
global ofs, size, scal
global bbox -- bounding box for display
global namearray = #()
-- global infofeed
-- create the section to output into ------------------------------------------
fn create_section = (	
	global sec = #() -- new array
	sec[size.x] = undefined
	for x = 1 to size.x do
	(
		sec[x] = #()
		sec[x][size.y] = undefined
		for y = 1 to size.y do
		(
			sec[x][y] = #()
			sec[x][y][size.z] = undefined
		)
	)
--sec
)

-- create a box given the top-left-corner and the size; coords are scaled -----
fn make_box atopleftcorner asize = (
	local csize = (asize * scal)
	local ccen = (csize/2) + ((atopleftcorner-1) * scal) + ofs
	box size:csize pos:ccen -- return it
)

-- create bounding box for display --------------------------------------------
fn create_display_box = (
	bbox = make_box [1,1,1] size
)


-- plot the section to screen; slow, uses a lot of ram ------------------------
fn plot_section = (
	-- prepare gui
	create_display_box()
	local voxels = 0
	local cell = box width:scal.x length:scal.y height:scal.z wireColor:bbox.wireColor parent:bbox
	local pos = ofs + (scal / 2)
	global vxlcreated = 0
	
		for x = 1 to size.x do (
		pos.y = ofs.y + (scal.y / 2)
		for y = 1 to size.y do (
			pos.z = ofs.z + (scal.z / 2)
			for z = 1 to size.z do (
				if sec[x][y][z] != undefined then 
				(
					cell.pos = pos
					(copy cell).wireColor = cell.wireColor -- copies don't seem to have the right colour?
					voxels += 1
					--xx = x as integer
					--yy = y as integer
					--zz = z as integer
					--test = "0" + ","+ xx as string + "," + yy as string + "," + zz as string + "," + "1" + "," + "50"
					--append  namearray test
				)
				
				pos.z += scal.z -- move to next z
			)
			pos.y += scal.y -- move to next y
		)
		
		pos.x += scal.x -- move to next x
		gc() -- run garbage collector now
	)
	format "% voxels plotted\n" voxels
	vxlcreated = voxels
	
)

-----------------------------------------------------------------------------------
fn plot_export = (
	-- prepare gui
	--create_display_box()
	local voxels = 0
	--local cell = box width:scal.x length:scal.y height:scal.z wireColor:bbox.wireColor parent:bbox
	local pos = ofs + (scal / 2)
	global xx, yy, zz, test
	global vxlcreated = 0
	
		for x = 1 to size.x do (
		pos.y = ofs.y + (scal.y / 2)
		for y = 1 to size.y do (
			pos.z = ofs.z + (scal.z / 2)
			for z = 1 to size.z do (
				if sec[x][y][z] != undefined then 
				(
					--cell.pos = pos
					--(copy cell).wireColor = cell.wireColor -- copies don't seem to have the right colour?
					voxels += 1
					xx = x as integer
					yy = y as integer
					zz = z as integer
					test = "0" + ","+ xx as string + "," + yy as string + "," + zz as string + "," + "1" + "," + newspincol as string --"50"
					append  namearray test
				)
				
				pos.z += scal.z -- move to next z
			)
			pos.y += scal.y -- move to next y
		)
		
		pos.x += scal.x -- move to next x
		--gc() -- run garbage collector now
	)
	format "% voxels plotted\n" voxels
	vxlcreated = voxels
)
---------------------------------------------------------------------------------------------


-- search on xy axis ----------------------------------------------------------
-- dir: 1 if from z=0 to size.z, -1 if from z=size.z downto 0
fn findxy node dir = (
	local sz, ez, zstep -- start-x, end-z, z-step
	local norm = [0,0,dir] -- direction to cast rays in
	local casts = 0, voxels = 0, outofbounds = 0
	local nray
	-- determine starting z from direction
	if dir == 1 then (
		sz = 0
		ez = size.z 
		zstep = scal.z
	) else (
		sz = size.z
		ez = 0
		zstep = -scal.z
	)
	format "searching xz from % to % with %..\n" sz ez norm
	-- do the search
	for x = 1 to size.x do
	(
		for y = 1 to size.y do
		(
			-- init ray to cast
			local spos, cray, nray, z
			spos = ([x - 0.5,y - 0.5 ,sz] * scal) + ofs
			cray = ray spos norm
			-- format "searching %\n" cray
			-- cast the first ray
			nray = intersectRay node cray
			casts += 1
			-- while hits
			while iskindof nray ray do (
				-- mark ray
				z = floor((nray.pos.z - ofs.z) / scal.z) + 1
				if (z >= 1.0) and (z <= size.z) then (
					sec[x][y][z] = 1 -- impossible colour, used to flag it
					voxels += 1
					
				) else (
					outofbounds += 1
					exit -- stop looking
				)
				-- find next
				spos += [0,0,z * zstep]
				cray = ray spos norm
				nray = intersectRay node cray
				casts += 1
			)
		)
	)
	format "% rays cast, % voxels found, % out-of-bounds\n" casts voxels outofbounds
)

-- search on xz axis ----------------------------------------------------------
-- dir: 1 if from z=0 to size.z, -1 if from z=size.z downto 0
fn findxz node dir = (
	local sy, ey, ystep -- start-y, end-y, y-step
	local norm = [0,dir,0] -- direction to cast rays in
	local casts = 0, voxels = 0, outofbounds = 0
	local nray
	-- determine starting y from direction
	if dir == 1 then (
		sy = 0
		ey = size.y 
		ystep = scal.y
	) else (
		sy = size.y
		ey = 0
		ystep = -scal.y
	)
	format "searching xz from % to % with %..\n" sy ey norm
	-- do the search
	for x = 1 to size.x do
	(
		for z = 1 to size.z do
		(
			-- init ray to cast

			local spos, cray, nray, y
			spos = ([x - 0.5,sy,z - 0.5] * scal) + ofs
			cray = ray spos norm
			-- format "searching %\n" cray
			-- cast the first ray
			nray = intersectRay node cray
			casts += 1
			-- while hits
			while iskindof nray ray do (
				-- mark ray
				y = floor((nray.pos.y - ofs.y) / scal.y) + 1
				if (y >= 1.0) and (y <= size.y) then (
					sec[x][y][z] = 1 -- impossible colour, used to flag it
					voxels += 1
				
				) else (
					outofbounds += 1
					exit -- stop looking
				)
				-- find next
				spos += [0,y * ystep,0]
				cray = ray spos norm
				nray = intersectRay node cray
				casts += 1
			)
		)
	)
	format "% rays cast, % voxels found, % out-of-bounds\n" casts voxels outofbounds
)

-- search on yz axis ----------------------------------------------------------
-- dir: 1 if from z=0 to size.z, -1 if from z=size.z downto 0
fn findyz node dir = (
	local sx, ex, xstep -- start-x, end-x, x-step
	local norm = [dir,0,0] -- direction to cast rays in
	local casts = 0, voxels = 0, outofbounds = 0
	--local vxlTime
	local nray
	-- determine starting y from direction
	if dir == 1 then (
		sx = 0
		ex = size.x 
		xstep = scal.x
	) else (
		sx = size.x
		ex = 0
		xstep = -scal.x
	)
	format "searching yz from % to % with %..\n" sx ex norm
	-- do the search
	for y = 1 to size.y do
	(			

		for z = 1 to size.z do
		(
			-- init ray to cast
			local spos, cray, nray, x
			spos = ([sx,y - 0.5,z - 0.5] * scal) + ofs
			cray = ray spos norm
			-- format "searching %\n" cray
			-- cast the first ray
			nray = intersectRay node cray
			casts += 1
			-- while hits
			while iskindof nray ray do (
				-- mark ray
				x = floor((nray.pos.x - ofs.x) / scal.x) + 1
				if (x >= 1.0) and (x <= size.x) then (
					sec[x][y][z] = 1 -- impossible colour, used to flag it
					voxels += 1
					
				) else (
					outofbounds += 1
					exit -- stop looking
				)
				-- find next
				spos += [x * xstep,0,0]
				cray = ray spos norm
				nray = intersectRay node cray
				casts += 1
			)
		)
	)
	format "% rays cast, % voxels found, % out-of-bounds\n" casts voxels outofbounds
	)




-- rollout routines -----------------------------------------------------------
--***************************************************************************
--***************************************************************************

-- rollout for about
--
rollout efx_rollout_voxel_about "About"
(
	label l1 "Voxel Generator - Geert Van der Poorten"
	--label l3 "by "
	--label l4 "Copyright 2002(c)"
	label l5 "Voxelengine - Will Sutton"
	--label l6 " by "
	label l2 "Version 1.02"

)

--
-- rollout for parameters
--


rollout efx_rollout_voxel_param "Parameters"
(
	local node, ext

	group "Object Information"
	(
		label node_name	across: 2
		label faces
		pickbutton pick_node "  PICK OBJECT  " width:100 height:30
		
	 	label dimx align:#left across: 3	
		label dimy align:#centre 
		label dimz align:#right
	)
	group "Info"
	(
		--label info1
		--label info2
		label top_R
		label front_R
		label side_R
	)
	
	group "Basic Voxelizing Color"
	(
	spinner spincol "change base color:      "range:[16,218,50] type:#integer  align:#right width:100 height:25 --across:2
	--colorpicker colbox align:#right
	)

--&	group "Get Texture Maps"
--	(
--	button do_rendertop " TOP Map " width:80 height:25 align:#left across:2 enabled:false
 --	label total_t align:#right 
--
--	button do_renderfront " FRONT Map " width:80 height:25 align:#left across:2 enabled:false
 --	label total_f align:#right 
--
--	button do_renderback " BACK Map " width:80 height:25 align:#left across:2 enabled:false
 --	label total_b align:#right 
--
--	button do_renderleft " LEFT Map " width:80 height:25  align:#left across:2 enabled:false
 --	label total_l align:#right 
--
--	button do_renderright " RIGHT Map " width:80 height:25 align:#left across:2 enabled:false
 --	label total_r align:#right 
--
--	label l4 "top/bottom maps are the same" enable:false
--	)

	group "Number of Voxels"
	(
		label vxlx across: 3
		label vxly
		label vxlz
		button do_create " VOXELIZE "	width:100 height:30
		label vxlt
		label vxlc
	
	ProgressBar vxlTime "Voxelizing" color:red width:200 height:12

	label infofeed " Stage Progress"
	label feedbck "Voxel Creation Time"

	)
	group "Export to TXT file"
	(
	edittext exptpath "Path"	text:"C:\\Westwood\\RA2\\MAX2VXL.txt" 
	label l5 "" 
	button do_export "  3DSMAX EXPORT "	width:100 height:30
		label labtxt01
		button do_gmaxexport "GMAX EXPORT "	width:100 height:30
		
	)



-- ---------------------------------------------------

on  do_rendertop pressed do
	(
	f1 = getopenFileName filename:"c:\\Program Files\\3dsmax3_1\\Images\\top.bmp" types:"Windows Bitmap (*.BMP)|*.bmp"
	top=openBitMap f1
	ht=top.height
	wt=top.width
	counter_t=0
		for r=0 to ht-1 do
		( 
			pixels=getpixels top [0,r] wt 		
				for c=1 to wt do 					
					(
						p=pixels[c] 					
							if 
								pixels[c] == color 0 0 255  
							then 
								counter_t = (counter_t )+ 1
					)
		)
	percent_t = counter_t*100/(wt*ht)
	total_t.text = counter_t as string + "    " + percent_t as string + " %"
	)


on  do_renderfront pressed do
	(
	f2 = getopenFileName filename:"c:\\Program Files\\3dsmax3_1\\Images\\front.bmp" types:"Windows Bitmap (*.BMP)|*.bmp"
	front=openBitMap f2
	hf=front.height
	wf=front.width
	counter_f=0
		for r=0 to hf-1 do
		( 
			pixels=getpixels front [0,r] wf 		
				for c=1 to wf do 					
					(
						p=pixels[c] 					
							if 
								pixels[c] == color 0 0 255  
							then 
								counter_f = (counter_f )+ 1
					)
		)
	percent_f = counter_f*100/(wf*hf)
	total_f.text = counter_f as string + "    " + percent_f as string + " %"
	)


on  do_renderback pressed do
	(
	f2 = getopenFileName filename:"c:\\Program Files\\3dsmax3_1\\Images\\back.bmp" types:"Windows Bitmap (*.BMP)|*.bmp"
	front=openBitMap f2
	hf=front.height
	wf=front.width
	counter_f=0
		for r=0 to hf-1 do
		( 
			pixels=getpixels front [0,r] wf 		
				for c=1 to wf do 					
					(
						p=pixels[c] 					
							if 
								pixels[c] == color 0 0 255  
							then 
								counter_f = (counter_f )+ 1
					)
		)
	percent_f = counter_f*100/(wf*hf)
	total_f.text = counter_f as string + "    " + percent_f as string + " %"
	)



on  do_renderleft pressed do
	(
	f3 = getopenFileName filename:"c:\\Program Files\\3dsmax3_1\\Images\\Left.bmp" types:"Windows Bitmap (*.BMP)|*.bmp"
	side=openBitMap f3
	hs=side.height
	ws=side.width
	counter_s=0
		for r=0 to hs-1 do
		( 
			pixels=getpixels side [0,r] ws 		
				for c=1 to ws do 					
					(
						p=pixels[c] 					
							if 
								pixels[c] == color 0 0 255  
							then 
								counter_s = (counter_s )+ 1
					)
		)
	percent_s = counter_s*100/(ws*hs)
	total_s.text = counter_s as string + "    " + percent_s as string + " %"
	)

on  do_renderright pressed do
	(
	f3 = getopenFileName filename:"c:\\Program Files\\3dsmax3_1\\Images\\Right.bmp" types:"Windows Bitmap (*.BMP)|*.bmp"
	side=openBitMap f3
	hs=side.height
	ws=side.width
	counter_s=0
		for r=0 to hs-1 do
		( 
			pixels=getpixels side [0,r] ws 		
				for c=1 to ws do 					
					(
						p=pixels[c] 					
							if 
								pixels[c] == color 0 0 255  
							then 
								counter_s = (counter_s )+ 1
					)
		)
	percent_s = counter_s*100/(ws*hs)
	total_s.text = counter_s as string + "    " + percent_s as string + " %"
	)

	on pick_node picked obj do
	(
		global obj
		vxlcreated = 0
		vxlTime.value=0 		--resettting progressbar
		feedbck.text = "Voxel Creation Time"
		infofeed.text = " Stage Progress"
		-- bounding box generation
		node = obj
		select obj
		pcount=getpolygoncount node
		polycount= pcount [1]
		sel =obj.name
		ext = node.max - node.min
		
	format "checking %\n" sel
	ofs = obj.min
	size = obj.max
	format "model bounds is % to %\n" ofs size
	size -= ofs
	scal = [10,10,10]
	size /= scal
	size.x = ceil(size.x)
	size.y = ceil(size.y)
	size.z = ceil(size.z)
	format "model size = %\n" size

		
		
		
		vxls = (ext/10)
		vxlsize =(vxls +.5) 
	Vxlsize.x = (vxls.x +.5) as integer
	Vxlsize.y = (vxls.y +.5) as integer
	Vxlsize.z = (vxls.z +.5) as integer
	Vxltot = Vxlsize.x * Vxlsize.y * Vxlsize.z
	x_as = (ext.x+.5) as integer
	y_as = (ext.y+.5) as integer
	z_as = (ext.z+.5) as integer
	-- information
		node_name.text = node.name
		dimx.text = "X: " + (ext.x as string)
		dimy.text = "Y: " + (ext.y as string)
		dimz.text = "Z: " + (ext.z as string)
		vxlx.text = "X: " + (vxlsize.x as string)
		vxly.text = "Y: " + (vxlsize.y as string)
		vxlz.text = "Z: " + (vxlsize.z as string)
		vxlt.text = "Voxel Grid Nodes : " + (Vxltot as string)
--		info1.text = "Use the rounded coordinates from above"
--		info2.text = "as numbers for the box-selected rendering"
		top_R.text = " TOP          "+"Width : " + ((vxls.x +.5) as integer) as string + "  Height : " + ((vxls.y +.5) as integer) as string
		front_R.text = "  FRONT    "+"Width : " + x_as as string + "  Height : " + z_as as string
		side_R.text = "SIDE        "+"Width : " + y_as as string + "  Height : " + z_as as string
		faces.text =  polycount as string + " Polygons"


)
-- ----------------------------------------------------






	on bounding_box_maker open do
	(
		node_name.text = "None"
	)

on spincol changed val do
(
newspincol = spincol.value
--selectedcolbox= case newspincol.state of
--	( 	0:
--		1:
--	)
--colbox = color selectedcolbox
)

	on do_create pressed do
	(
	
	--start = timeStamp() 

		if node != undefined then
		(
			escapeEnable = false
			--progressStart "Processing:"
			undo on
			(
				--draw_voxel node (vxlsize.x)(vxlsize.y)(vxlsize.z)
				--status.text = ""

		--max/gmax toggle
		if 
		(
		(maxversion())[2]==128
		)
		then 
		(
		do_export.enabled=false
		labtxt01.text = "Renderer not available in GMAX"
		)
		else
		(
		do_gmaxexport.enabled=false
		)



	local starttime = timestamp()
	-- create section
	format "Creating empty section\n"
	create_section()
--	sec=create_section()
	--infofeed.text =  "section size is [ "+  sec.count as string +","+ sec[1].count as string  +","+ sec[1][1].count as string   +" ]"  
	-- do the search
	local node = copy $ -- must use copy
	-- find on all three axis in both directions
infofeed.text = "Voxelizing process started"
vxlTime.value=0
	findxy node  1
vxlTime.value += ceil(100/6+.5)
	findxy node -1
vxlTime.value += ceil(100/6+.5)
	findxz node  1
vxlTime.value += ceil(100/6+.5)
	findxz node -1
vxlTime.value += ceil(100/6+.5)
	findyz node  1
vxlTime.value += ceil(100/6+.5)
	findyz node -1
vxlTime.value += ceil(100/6+.5)
	delete node
feedbck.text = "Voxelizing took " + ((timestamp() - starttime) / 1000.0) as string  +" seconds"  --format "Processing took % seconds\n" ((timestamp() - starttime) / 1000.0) 
	-- plot for display
	starttime = timestamp()
	--plot_section()
	infofeed.text = "Voxels ready for export"
	plot_export ()
	-- done
	format "cleaning up ram\n"
	sec = undefined
			)

		--	progressEnd()
			escapeEnable = true
		)
		vxlc.text = "Created Voxels : " + (vxlcreated as string)
	)







	on do_export pressed do
	(

f= createfile exptpath.text
for i=1 to vxlcreated do format "%\n" namearray[i] to:f
close f
	)

	on do_gmaxexport pressed do
	(
clearListener() 
for i=1 to vxlcreated do format "%\n" namearray[i] --to:f
messagebox "GMAX has NO export possibilities. Copy/paste the data from the Listener (F11)"
	)




)




rollout rollout_text_param "Texturing parameters"
(
)


-- floater routines -----------------------------------------------------------


if efx_floater_voxel != undefined then closeRolloutFloater efx_floater_voxel

efx_floater_voxel = newRolloutFloater "Voxel Generator" 250 710



addRollout efx_rollout_voxel_param efx_floater_voxel
--addRollout rollout_text_param efx_floater_voxel
addRollout efx_rollout_voxel_about efx_floater_voxel rolledUp:true
